import { Operations, Operation } from './operation';
import { JsonPatchError } from './json-patch-error';
import { arrOps, objOps } from './operation-extensions';
import { areEquals } from './deep-equals';
var Helpers = /** @class */ (function () {
    function Helpers() {
    }
    /**
     * Deeply clone the object.
     * https://jsperf.com/deep-copy-vs-json-stringify-json-parse/25 (recursiveDeepCopy)
     * @param  {any} obj value to clone
     * @return {any} cloned obj
     */
    Helpers.deepClone = function (obj) {
        switch (typeof obj) {
            case 'object':
                // Faster than ES5 clone - http://jsperf.com/deep-cloning-of-objects/5
                return JSON.parse(JSON.stringify(obj));
            case 'undefined':
                // this is how JSON.stringify behaves for array items
                return null;
            default:
                // no need to clone primitives
                return obj;
        }
    };
    /**
     * Retrieves a value from a JSON document by a JSON pointer.
     * Returns the value.
     *
     * @param document The document to get the value from
     * @param pointer an escaped JSON pointer
     * @return The retrieved value
     */
    Helpers.getValueByPointer = function (document, pointer) {
        if (pointer === '') {
            return document;
        }
        var getOriginalDestination = new Operation(Operations.Get, pointer);
        this.applyOperation(document, getOriginalDestination);
        return getOriginalDestination.value;
    };
    /**
     * Apply a single JSON Patch Operation on a JSON document.
     * Returns the {newDocument, result} of the operation.
     * It modifies the `document` and `operation` objects - it gets the values by reference.
     * If you would like to avoid touching your values, clone them:
     * `jsonpatch.applyOperation(document, jsonpatch._deepClone(operation))`.
     *
     * @param document The document to patch
     * @param operation The operation to apply
     * @param validateOperation `false` is without validation, `true` to use default jsonpatch's validation, or you can pass a `validateOperation` callback to be used for validation.
     * @param mutateDocument Whether to mutate the original document or clone it before applying
     * @return `{newDocument, result}` after the operation
     */
    Helpers.applyOperation = function (document, operation, validateOperation, mutateDocument) {
        if (validateOperation === void 0) { validateOperation = false; }
        if (mutateDocument === void 0) { mutateDocument = true; }
        if (validateOperation) {
            if (typeof validateOperation === 'function') {
                validateOperation(operation, 0, document, operation.path);
            }
            else {
                Helpers.validator(operation, 0);
            }
        }
        /* ROOT OPERATIONS */
        if (operation.path === '') {
            var returnValue = { newDocument: document };
            if (operation.op === 'add') {
                returnValue.newDocument = operation.value;
                return returnValue;
            }
            else if (operation.op === 'replace') {
                returnValue.newDocument = operation.value;
                returnValue.removed = document; // document we removed
                return returnValue;
            }
            else if (operation.op === 'move' || operation.op === 'copy') {
                returnValue.newDocument = Helpers.getValueByPointer(document, operation.from); // get the value by json-pointer in `from` field
                if (operation.op === 'move') {
                    returnValue.removed = document;
                }
                return returnValue;
            }
            else if (operation.op === 'test') {
                returnValue.test = areEquals(document, operation.value);
                if (returnValue.test === false) {
                    throw new JsonPatchError('Test operation failed', 'TEST_OPERATION_FAILED', 0, operation, document);
                }
                returnValue.newDocument = document;
                return returnValue;
            }
            else if (operation.op === 'remove') {
                returnValue.removed = document;
                returnValue.newDocument = null;
                return returnValue;
            }
            else if (operation.op === '_get') {
                operation.value = document;
                return returnValue;
            }
            else {
                if (validateOperation) {
                    throw new JsonPatchError('Operation `op` property is not one of operations defined in RFC-6902', 'OPERATION_OP_INVALID', 0, operation, document);
                }
                else {
                    return returnValue;
                }
            }
            /* END ROOT OPERATIONS */
        }
        else {
            if (!mutateDocument) {
                document = Helpers.deepClone(document);
            }
            var path = operation.path || '';
            var keys = path.split('/');
            var obj = document;
            var t = 1; // skip empty element - http://jsperf.com/to-shift-or-not-to-shift
            var len = keys.length;
            var existingPathFragment = void 0;
            var key = void 0;
            var validateFunction = void 0;
            if (typeof validateOperation === 'function') {
                validateFunction = validateOperation;
            }
            else {
                validateFunction = Helpers.validator;
            }
            while (true) {
                key = keys[t];
                if (validateOperation) {
                    if (existingPathFragment) {
                        if (obj[key] === undefined) {
                            existingPathFragment = keys.slice(0, t).join('/');
                        }
                        else if (t === len - 1) {
                            existingPathFragment = operation.path;
                        }
                        if (existingPathFragment !== undefined) {
                            validateFunction(operation, 0, document, existingPathFragment);
                        }
                    }
                }
                t++;
                if (Array.isArray(obj)) {
                    if (key === '-') {
                        key = obj.length;
                    }
                    else {
                        if (validateOperation && !Helpers.isInteger(key)) {
                            throw new JsonPatchError('Expected an unsigned base-10 integer value, making the new referenced value the array element with the zero-based index', 'OPERATION_PATH_ILLEGAL_ARRAY_INDEX', 0, operation.path, operation);
                        }
                        else if (Helpers.isInteger(key)) {
                            // only parse key when it's an integer for `arr.prop` to work
                            // tslint:disable-next-line:no-bitwise
                            key = ~~key;
                        }
                    }
                    if (t >= len) {
                        if (validateOperation && operation.op === 'add' && key > obj.length) {
                            throw new JsonPatchError('The specified index MUST NOT be greater than the number of elements in the array', 'OPERATION_VALUE_OUT_OF_BOUNDS', 0, operation.path, operation);
                        }
                        var returnValue = arrOps[operation.op].call(operation, obj, key, document); // Apply patch
                        if (returnValue.test === false) {
                            throw new JsonPatchError('Test operation failed', 'TEST_OPERATION_FAILED', 0, operation, document);
                        }
                        return returnValue;
                    }
                }
                else {
                    if (key && key.indexOf('~') !== -1) {
                        key = Helpers.unescapePathComponent(key);
                    }
                    if (t >= len) {
                        var returnValue = objOps[operation.op].call(operation, obj, key, document); // Apply patch
                        if (returnValue.test === false) {
                            throw new JsonPatchError('Test operation failed', 'TEST_OPERATION_FAILED', 0, operation, document);
                        }
                        return returnValue;
                    }
                }
                obj = obj[key];
            }
        }
    };
    // 3x faster than cached /^\d+$/.test(str)
    Helpers.isInteger = function (str) {
        var i = 0;
        var len = str.length;
        var charCode;
        while (i < len) {
            charCode = str.charCodeAt(i);
            if (charCode >= 48 && charCode <= 57) {
                i++;
                continue;
            }
            return false;
        }
        return true;
    };
    Helpers.generate = function (mirror, obj, patches, path) {
        if (obj === mirror) {
            return;
        }
        if (typeof obj.toJSON === 'function') {
            obj = obj.toJSON();
        }
        var newKeys = this.objectKeys(obj);
        var oldKeys = this.objectKeys(mirror);
        var changed = false;
        var deleted = false;
        // if ever "move" operation is implemented here, make sure this test runs OK: "should not generate the same patch twice (move)"
        for (var okl = oldKeys.length - 1; okl >= 0; okl--) {
            var oldKey = oldKeys[okl];
            var oldVal = mirror[oldKey];
            if (Helpers.hasOwnProperty(obj, oldKey) && !(obj[oldKey] === undefined && oldVal !== undefined && Array.isArray(obj) === false)) {
                var newVal = obj[oldKey];
                if (typeof oldVal === 'object' && oldVal != null && typeof newVal === 'object' && newVal != null) {
                    Helpers.generate(oldVal, newVal, patches, path + '/' + Helpers.escapePathComponent(oldKey));
                }
                else {
                    if (oldVal !== newVal) {
                        changed = true;
                        patches.push({ op: 'replace', path: path + '/' + Helpers.escapePathComponent(oldKey), value: Helpers.deepClone(newVal) });
                    }
                }
            }
            else {
                patches.push({ op: 'remove', path: path + '/' + Helpers.escapePathComponent(oldKey) });
                deleted = true; // property has been deleted
            }
        }
        if (!deleted && newKeys.length === oldKeys.length) {
            return;
        }
        for (var _i = 0, newKeys_1 = newKeys; _i < newKeys_1.length; _i++) {
            var newKey = newKeys_1[_i];
            if (!Helpers.hasOwnProperty(mirror, newKey) && obj[newKey] !== undefined) {
                patches.push({ op: 'add', path: path + '/' + Helpers.escapePathComponent(newKey), value: Helpers.deepClone(obj[newKey]) });
            }
        }
    };
    /**
     * Apply a full JSON Patch array on a JSON document.
     * Returns the {newDocument, result} of the patch.
     * It modifies the `document` object and `patch` - it gets the values by reference.
     * If you would like to avoid touching your values, clone them:
     * `jsonpatch.applyPatch(document, jsonpatch._deepClone(patch))`.
     *
     * @param document The document to patch
     * @param patch The patch to apply
     * @param validateOperation `false` is without validation, `true` to use default jsonpatch's validation, or you can pass a `validateOperation` callback to be used for validation.
     * @param mutateDocument Whether to mutate the original document or clone it before applying
     * @return An array of `{newDocument, result}` after the patch
     */
    Helpers.applyPatch = function (document, patch, validateOperation, mutateDocument) {
        if (mutateDocument === void 0) { mutateDocument = true; }
        if (validateOperation) {
            if (!Array.isArray(patch)) {
                throw new JsonPatchError('Patch sequence must be an array', 'SEQUENCE_NOT_AN_ARRAY');
            }
        }
        if (!mutateDocument) {
            document = Helpers.deepClone(document);
        }
        var results = new Array(patch.length);
        for (var i = 0, length_1 = patch.length; i < length_1; i++) {
            results[i] = Helpers.applyOperation(document, patch[i], validateOperation);
            document = results[i].newDocument; // in case root was replaced
        }
        results.newDocument = document;
        return results;
    };
    /**
     * Validates a single operation. Called from `jsonpatch.validate`. Throws `JsonPatchError` in case of an error.
     * @param {object} operation - operation object (patch)
     * @param {number} index - index of operation in the sequence
     * @param {object} [document] - object where the operation is supposed to be applied
     * @param {string} [existingPathFragment] - comes along with `document`
     */
    Helpers.validator = function (operation, index, document, existingPathFragment) {
        if (typeof operation !== 'object' || operation === null || Array.isArray(operation)) {
            throw new JsonPatchError('Operation is not an object', 'OPERATION_NOT_AN_OBJECT', index, operation, document);
        }
        else if (!objOps[operation.op]) {
            throw new JsonPatchError('Operation `op` property is not one of operations defined in RFC-6902', 'OPERATION_OP_INVALID', index, operation, document);
        }
        else if (typeof operation.path !== 'string') {
            throw new JsonPatchError('Operation `path` property is not a string', 'OPERATION_PATH_INVALID', index, operation, document);
        }
        else if (operation.path.indexOf('/') !== 0 && operation.path.length > 0) {
            // paths that aren't empty string should start with "/"
            throw new JsonPatchError('Operation `path` property must start with "/"', 'OPERATION_PATH_INVALID', index, operation, document);
        }
        else if ((operation.op === 'move' || operation.op === 'copy') && typeof operation.from !== 'string') {
            throw new JsonPatchError('Operation `from` property is not present (applicable in `move` and `copy` operations)', 'OPERATION_FROM_REQUIRED', index, operation, document);
        }
        else if ((operation.op === 'add' || operation.op === 'replace' || operation.op === 'test') && operation.value === undefined) {
            throw new JsonPatchError('Operation `value` property is not present (applicable in `add`, `replace` and `test` operations)', 'OPERATION_VALUE_REQUIRED', index, operation, document);
        }
        else if ((operation.op === 'add' || operation.op === 'replace' || operation.op === 'test') && this.hasUndefined(operation.value)) {
            throw new JsonPatchError('Operation `value` property is not present (applicable in `add`, `replace` and `test` operations)', 'OPERATION_VALUE_CANNOT_CONTAIN_UNDEFINED', index, operation, document);
        }
        else if (document) {
            if (operation.op === 'add') {
                var pathLen = operation.path.split('/').length;
                var existingPathLen = existingPathFragment.split('/').length;
                if (pathLen !== existingPathLen + 1 && pathLen !== existingPathLen) {
                    throw new JsonPatchError('Cannot perform an `add` operation at the desired path', 'OPERATION_PATH_CANNOT_ADD', index, operation, document);
                }
            }
            else if (operation.op === 'replace' || operation.op === 'remove' || operation.op === '_get') {
                if (operation.path !== existingPathFragment) {
                    throw new JsonPatchError('Cannot perform the operation at a path that does not exist', 'OPERATION_PATH_UNRESOLVABLE', index, operation, document);
                }
            }
            else if (operation.op === 'move' || operation.op === 'copy') {
                var existingValue = new Operation(Operations.Get, operation.from, undefined);
                var error = Helpers.validate([existingValue], document);
                if (error && error.name === 'OPERATION_PATH_UNRESOLVABLE') {
                    throw new JsonPatchError('Cannot perform the operation from a path that does not exist', 'OPERATION_FROM_UNRESOLVABLE', index, operation, document);
                }
            }
        }
    };
    /**
     * Validates a sequence of operations. If `document` parameter is provided, the sequence is additionally validated against the object document.
     * If error is encountered, returns a JsonPatchError object
     * @param sequence
     * @param document
     * @returns {JsonPatchError|undefined}
     */
    Helpers.validate = function (sequence, document, externalValidator) {
        try {
            if (!Array.isArray(sequence)) {
                throw new JsonPatchError('Patch sequence must be an array', 'SEQUENCE_NOT_AN_ARRAY');
            }
            if (document) {
                // clone document and sequence so that we can safely try applying operations
                Helpers.applyPatch(Helpers.deepClone(document), Helpers.deepClone(sequence), externalValidator || true);
            }
            else {
                var validatorFunction = externalValidator.validate || Helpers.validator;
                for (var i = 0; i < sequence.length; i++) {
                    validatorFunction(sequence[i], i, document, undefined);
                }
            }
        }
        catch (e) {
            if (e instanceof JsonPatchError) {
                return e;
            }
            else {
                throw e;
            }
        }
    };
    /**
     * Recursively checks whether an object has any undefined values inside.
     */
    Helpers.hasUndefined = function (obj) {
        if (obj === undefined) {
            return true;
        }
        if (obj) {
            if (Array.isArray(obj)) {
                for (var a = 0, len = obj.length; a < len; a++) {
                    if (this.hasUndefined(obj[a])) {
                        return true;
                    }
                }
            }
            else if (typeof obj === 'object') {
                var objKeys = Helpers.objectKeys(obj);
                var objKeysLength = objKeys.length;
                for (var i = 0; i < objKeysLength; i++) {
                    if (Helpers.hasUndefined(obj[objKeys[i]])) {
                        return true;
                    }
                }
            }
        }
        return false;
    };
    Helpers.hasOwnProperty = function (obj, key) {
        return Helpers._hasOwnProperty.call(obj, key);
    };
    Helpers.objectKeys = function (obj) {
        if (Array.isArray(obj)) {
            var arrayKeys = new Array(obj.length);
            for (var k = 0; k < arrayKeys.length; k++) {
                arrayKeys[k] = '' + k;
            }
            return arrayKeys;
        }
        if (Object.keys) {
            return Object.keys(obj);
        }
        var keys = [];
        for (var i in obj) {
            if (Helpers.hasOwnProperty(obj, i)) {
                keys.push(i);
            }
        }
        return keys;
    };
    /**
     * Unescapes a json pointer path
     * @param path The escaped pointer
     * @return The unescaped path
     */
    Helpers.unescapePathComponent = function (path) {
        var newPath = path.replace(/~1/g, '/');
        newPath = newPath.replace(/~0/g, '~');
        return newPath;
    };
    /**
     * Escapes a json pointer path
     * @param path The raw pointer
     * @return the Escaped path
     */
    Helpers.escapePathComponent = function (path) {
        if (path.indexOf('/') === -1 && path.indexOf('~') === -1) {
            return path;
        }
        return path.replace(/~/g, '~0').replace(/\//g, '~1');
    };
    Helpers._hasOwnProperty = Object.prototype.hasOwnProperty;
    return Helpers;
}());
export { Helpers };
//# sourceMappingURL=helpers.js.map