import { Helpers } from './helpers';
import { Operation, Operations } from './operation';
import { areEquals } from './deep-equals';
/* We use a Javascript hash to store each
 function. Each hash entry (property) uses
 the operation identifiers specified in rfc6902.
 In this way, we can map each patch operation
 to its dedicated function in efficient way.
 */
/* The operations applicable to an object */
export var objOps = {
    this: this,
    add: function (obj, key, document) {
        obj[key] = objOps.this.value;
        return { newDocument: document };
    },
    remove: function (obj, key, document) {
        var removed = obj[key];
        delete obj[key];
        return { newDocument: document, removed: removed };
    },
    replace: function (obj, key, document) {
        var removed = obj[key];
        obj[key] = objOps.this.value;
        return { newDocument: document, removed: removed };
    },
    move: function (obj, key, document) {
        /* in case move target overwrites an existing value,
        return the removed value, this can be taxing performance-wise,
        and is potentially unneeded */
        var removed = Helpers.getValueByPointer(document, objOps.this.path);
        if (removed) {
            removed = Helpers.deepClone(removed);
        }
        var removeOperation = new Operation(Operations.Remove, objOps.this.from);
        var originalValue = Helpers.applyOperation(document, removeOperation).removed;
        var addOperation = new Operation(Operations.Add, objOps.this.path, originalValue);
        Helpers.applyOperation(document, addOperation);
        return { newDocument: document, removed: removed };
    },
    copy: function (obj, key, document) {
        var valueToCopy = Helpers.getValueByPointer(document, objOps.this.from);
        // enforce copy by value so further operations don't affect source (see issue #177)
        var addOperation = new Operation(Operations.Add, objOps.this.path, Helpers.deepClone(valueToCopy));
        Helpers.applyOperation(document, addOperation);
        return { newDocument: document };
    },
    test: function (obj, key, document) {
        return { newDocument: document, test: areEquals(obj[key], objOps.this.value) };
    },
    _get: function (obj, key, document) {
        objOps.this.value = obj[key];
        return { newDocument: document };
    }
};
/* The operations applicable to an array. Many are the same as for the object */
export var arrOps = {
    this: this,
    add: function (arr, i, document) {
        if (Helpers.isInteger(i)) {
            arr.splice(i, 0, arrOps.this.value);
        }
        else {
            arr[i] = arrOps.this.value;
        }
        // this may be needed when using '-' in an array
        return { newDocument: document, index: i };
    },
    remove: function (arr, i, document) {
        var removedList = arr.splice(i, 1);
        return { newDocument: document, removed: removedList[0] };
    },
    replace: function (arr, i, document) {
        var removed = arr[i];
        arr[i] = arrOps.this.value;
        return { newDocument: document, removed: removed };
    },
    move: objOps.move,
    copy: objOps.copy,
    test: objOps.test,
    _get: objOps._get
};
//# sourceMappingURL=operation-extensions.js.map